import argparse


def str2bool(v) -> bool:
    """ Converts command line boolean statements to python boolean.

    Args
    ---
    - v: A string representing the command line boolean statements.

    Returns
    ---
    A python boolean.

    Raises
    ---
    - `v` does not match any command line boolean statements.
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')



import os
from typing import List


def read_char_map(char_map_filepath) -> List[str]:
    """ Read characters from `char_map_filepath` into a python list.

    Args
    ---
    char_map_filepath: A string representing path to the characters map file.

    Returns
    ---
    A python list with an extra "UNK" char for unkown character.

    Raises
    ---
    `char_map_filepath` does not exist.

    """
    assert os.path.exists, "char_map_filepath does not exist."

    with open(char_map_filepath, "r") as file:
        char_map = [i.strip("\n").split("\t")[0] for i in file.readlines()]
        char_map += ["UNK"]
        return char_map



import os
import numpy as np
from .network import Network
from .utils import read_pos_map, read_char_map, read_config

dir_name = os.path.dirname(os.path.abspath(__file__))
print(os.path.join(dir_name, 'configs/original_paper.json'))
config = read_config(os.path.join(dir_name, 'configs/original_paper.json'))
char_map = read_char_map(os.path.join(dir_name, 'pretrained/char_map.txt'))
pos_map = read_pos_map(os.path.join(dir_name, 'pretrained/pos_map.txt'))
weight_path = os.path.join(dir_name, 'pretrained/original_paper.weights.h5')

num_chars = len(char_map)
pos_to_index = {pos: i for i, pos in enumerate(pos_map)}
index_to_pos = {i: pos for i, pos in enumerate(pos_map)}
char_to_index = {char: i for i, char in enumerate(char_map)}

model = Network(
    output_dim=len(pos_map),
    embedding_dim=len(char_map),
    num_stacks=config["model"]["num_stacks"],
    hidden_layers_dim=config["model"]["hidden_layers_dim"],
    max_sentence_length=config["model"]["max_sentence_length"],
)
model.load_weights(weight_path, by_name=True)

def preprocess_sample(sentence):
    sentence_input_vector = np.zeros((config["model"]["max_sentence_length"], num_chars))
    for i, char in enumerate(sentence):
        if char in char_to_index:
            char_index = char_to_index[char]
        else:
            char_index = char_to_index["UNK"]
        sentence_input_vector[i, char_index] = 1

    return sentence_input_vector, sentence

def tokenize(samples):
    result = []
    for sample in samples:
        sentence_input_vector, sentence = preprocess_sample(sample)
        pred = model.predict(np.array([sentence_input_vector]))[0]

        words, tmp, pos = [], [], []
        for char_idx, pos_vector in enumerate(pred):
            if char_idx < len(sentence):
                pos_index_pred = np.argmax(pos_vector)

                if index_to_pos[pos_index_pred] == "NS":
                    tmp.append(sentence[char_idx])
                else:
                    if len(tmp) > 0:
                        words.append("".join(tmp))
                        tmp = []
                    pos.append(index_to_pos[pos_index_pred])
                    tmp.append(sentence[char_idx])

        if len(tmp) > 0:
            words.append("".join(tmp))
        result.append((words, pos))

    return result




import os
import argparse

parser = argparse.ArgumentParser(description='Create the char_map.txt and pos_map.txt file.')
parser.add_argument('--output_dir', type=str,
                    help='The directory to output the results.', default="output")
args = parser.parse_args()

os.makedirs(args.output_dir, exist_ok=True)

SEPARATOR = set("\u200b")
CONSTS = set(u"កខគឃងចឆជឈញដឋឌឍណតថទធនបផពភមយរលវឝឞសហឡអឣឤឥឦឧឨឩឪឫឬឭឮឯឰឱឲឳ")
VOWELS = set(u"឴឵ាិីឹឺុូួើឿៀេែៃោៅ\u17c6\u17c7\u17c8")
SUB = set(u"្")
DIAC = set(u"\u17c9\u17ca\u17cb\u17cc\u17cd\u17ce\u17cf\u17d0")
SYMS = set("៕។៛ៗ៚៙៘,.?!@#%^&*()_+-=[]/\<>")
NUMBERS = set(u"០១២៣៤៥៦៧៨៩0123456789")

characters = sorted(list(SEPARATOR) + list(CONSTS) + list(VOWELS) + list(SUB) + list(DIAC) + list(SYMS) + list(NUMBERS))

with open(os.path.join(args.output_dir, "char_map.txt"), "w") as file:
    for i, char in enumerate(characters):
        file.write(f"{char}\t{i}")
        if i != len(characters)-1:
            file.write("\n")

with open(os.path.join(args.output_dir, "pos_map.txt"), "w") as file:
    tags = ["AB", "AUX", "CC", "CD", "DT", "IN", "JJ", "VB", "NN", "PN", "PA", "PRO", "QT", "RB", "SYM", "NS"]
    for i, tag in enumerate(tags):
        file.write(f"{tag}")
        if i != len(tags)-1:
            file.write("\n")



from tensorflow.keras.models import Model
from tensorflow.keras.layers import LSTM, Dense, Bidirectional, Activation, Input, Flatten


def Network(output_dim, embedding_dim, num_stacks, hidden_layers_dim, max_sentence_length=None) -> Model:
    """ Defines the structure of the network.

    Args
    ---
    - output_dim: An int representing the size of the output vector.
    - embedding_dim: An int representing the size of the character vector.
    - num_stacks: An int representing the number of LSTM stacks.
    - hidden_layers_dim: An int representing the number of units for each hidden layer.
    - max_sentence_length: An int representing the max length of a sentence. (Defaults to None)

    Returns
    ---
    Keras model object.

    Raises
    ---
    - `output_dim` must be larger than 0
    - `embedding_dim` must be larger than 0
    - `num_stacks` must be larger than 0
    - `hidden_layers_dim` must be larger than 0

    References
    ---
    - Buoy, R., Taing, N., & Kor, S. (2021). Joint Khmer Word Segmentation and Part-of-Speech Tagging Using Deep Learning. arXiv preprint arXiv:2103.16801.
    - Loem, M. (2021, May 4). Joint Khmer Word Segmentation and POS tagging. Medium. Retrieved February 22, 2022, from https://towardsdatascience.com/joint-khmer-word-segmentation-and-pos-tagging-cad650e78d30
    """
    assert output_dim > 0, "output_dim must be larger than 0"
    assert embedding_dim > 0, "embedding_dim must be larger than 0"
    assert num_stacks > 0, "num_stacks must be larger than 0"
    assert hidden_layers_dim > 0, "hidden_layers_dim must be larger than 0"

    input_layer = Input(shape=(max_sentence_length, embedding_dim))
    x = Bidirectional(LSTM(hidden_layers_dim, return_sequences=True))(input_layer)
    for i in range(num_stacks - 1):
        x = Bidirectional(LSTM(hidden_layers_dim, return_sequences=i != num_stacks-1))(x)
    x = Dense(output_dim)(x)
    x = Activation("softmax")(x)
    model = Model(inputs=input_layer, outputs=x)
    return model



import os
import json
from typing import Dict


def read_config(config_filepath) -> Dict[str, str]:
    """ Read characters from `filepath` into a python list.

    Args
    ---
    config_filepath: A string representing path to the json config file.

    Returns
    ---
    A python dict representing the json config file.

    Raises
    ---
    - `config_filepath` does not exist.
    - `batch_size` must be larger than 0.
    - `learning_rate` must be larger than 0.

    """
    assert os.path.exists(config_filepath), "config_filepath does not exist."
    with open(config_filepath, "r") as config_file:
        config = json.load(config_file)
    assert config["training"]["batch_size"] > 0, "batch_size must be larger than 0"
    assert config["training"]["learning_rate"] > 0, "learning_rate must be larger than 0"

    return config



import os
from typing import List


def read_pos_map(pos_map_filepath) -> List[str]:
    """ Read POS tags from `pos_map_filepath` into a python list.

    Args
    ---
    pos_map_filepath: A string representing path to the pos map file.

    Returns
    ---
    A python list.

    Raises
    ---
    `pos_map_filepath` does not exist.

    """
    assert os.path.exists, "pos_map_filepath does not exist."
    with open(pos_map_filepath, "r") as file:
        pos_map = [i.strip("\n").split("\t")[0] for i in file.readlines()]
        return pos_map



import os
from typing import List


def read_samples(split_filepath) -> List[str]:
    """ Read samples from `split_filepath` into a python list.

    Args
    ---
    split_filepath: A string representing path to the split_filepath file.

    Returns
    ---
    A python list.

    Raises
    ---
    `split_filepath` does not exist.
    """
    assert os.path.exists(split_filepath), "split_filepath does not exist."

    with open(split_filepath, "r") as dataset_file:
        samples = [i.strip("\n").strip("") for i in dataset_file.readlines()]
        return samples



import os
import argparse
from str2bool import str2bool
from read_pos_map import read_pos_map
from read_char_map import read_char_map

parser = argparse.ArgumentParser(description='Prepare khPOS dataset to correct format.')
parser.add_argument('data_dir', type=str, help='Path to dataset directory.')
parser.add_argument('--split_sentences', type=str2bool, nargs='?',
                    help='whether to generate more samples by splitting the sentences into chunks of 0 to length.', default=False)
parser.add_argument('--output_dir', type=str,
                    help='The directory to output the results.', default="output")
args = parser.parse_args()

assert os.path.exists(args.data_dir), "data_dir does not exist"
os.makedirs(args.output_dir, exist_ok=True)

REVISED_TAG = {
    "M": "NN",
    "RPN": "PRO",
    "CUR": "SYM",
    "DBL": "SYM",
    "ETC": "SYM",
    "KAN": "SYM",
    "UH": "PA",
    "VB_JJ": "VB",
    "VCOM": "VB"
}


def process_line(line):
    X = ""
    y = ""
    chunks = line.strip("\n").split(" ")
    for chunk in chunks:
        word, tag = chunk.split("/")
        if tag in REVISED_TAG:
            tag = REVISED_TAG[tag]

        X += word

        y += f"/{tag}"

        for _ in word[1:]:
            y += "/NS"

    return X, y


def write_samples_to_file(input_filepath, output_filepath, split_sentences=False):
    writer = open(output_filepath, "w")
    count = 0
    MAX_SENTENCE_LENGTH = -999
    with open(input_filepath, "r") as data_file:
        lines = data_file.readlines()
        for i, line in enumerate(lines):
            if split_sentences:
                chunks = line.strip("\n").split(" ")
                for start in range(0, len(chunks)+1):
                    for end in range(start+1, len(chunks)+1):
                        sub_line = " ".join(chunks[start:end])
                        X, y = process_line(sub_line)
                        writer.write(X + "\t" + y + "\n")
                        count += 1

            X, y = process_line(line)
            if MAX_SENTENCE_LENGTH < len(X):
                MAX_SENTENCE_LENGTH = len(X)
            writer.write(X + "\t" + y + ("\n" if i != len(lines)-1 else ""))
            count += 1

    writer.close()
    print(f"written samples: {count} | MAX_SENTENCE_LENGTH: {MAX_SENTENCE_LENGTH}")


if __name__ == "__main__":
    training_file = os.path.join(args.data_dir, "corpus-draft-ver-1.0/data/after-replace/train.all2")
    test_file_open = os.path.join(args.data_dir, "corpus-draft-ver-1.0/data/OPEN-TEST")
    test_file_close = os.path.join(args.data_dir, "corpus-draft-ver-1.0/data/CLOSE-TEST")

    write_samples_to_file(training_file, os.path.join(args.output_dir, "train.txt"), args.split_sentences)
    write_samples_to_file(test_file_open, os.path.join(args.output_dir, "test_open.txt"))
    write_samples_to_file(test_file_close, os.path.join(args.output_dir, "test_close.txt"))
